#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     This is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#
"""Module for generating basemaps from various providers."""

import argparse
import concurrent.futures
import logging
import queue
import re
import sys
import threading
from pathlib import Path
from typing import Union

import geojson
import mercantile
from cpuinfo import get_cpu_info
from pmtiles.tile import (
    Compression as PMTileCompression,
)
from pmtiles.tile import (
    TileType as PMTileType,
)
from pmtiles.tile import (
    zxy_to_tileid,
)
from pmtiles.writer import Writer as PMTileWriter
from pySmartDL import SmartDL
from shapely.geometry import shape
from shapely.ops import unary_union

from osm_fieldwork.sqlite import DataFile, MapTile
from osm_fieldwork.xlsforms import xlsforms_path
from osm_fieldwork.yamlfile import YamlFile

# Instantiate logger
log = logging.getLogger(__name__)


def dlthread(
    dest: str,
    mirrors: list,
    tiles: list,
    xy: bool,
):
    """Thread to handle downloads for Queue.

    Args:
        dest (str): The filespec of the tile cache
        mirrors (list): The list of mirrors to get imagery
        tiles (list): The list of tiles to download
        xy (bool): Whether to swap the X & Y fields in the TMS URL
    """
    if len(tiles) == 0:
        # epdb.st()
        return
    # counter = -1

    # start = datetime.now()

    # totaltime = 0.0
    log.info("Downloading %d tiles in thread %d to %s" % (len(tiles), threading.get_ident(), dest))
    for tile in tiles:
        filespec = f"{tile[2]}/{tile[1]}/{tile[0]}"
        for site in mirrors:
            if site["source"] != "topo":
                filespec += "." + site["suffix"]
            url = site["url"]
            if site["source"] == "bing":
                bingkey = mercantile.quadkey(tile)
                remote = url % bingkey
            elif site["source"] == "google":
                path = f"x={tile[0]}&s=&y={tile[1]}&z={tile[2]}"
                remote = url % path
            elif site["source"] == "custom":
                if not xy:
                    # z/y/x format
                    path = f"{tile[2]}/{tile[1]}/{tile[0]}"
                else:
                    # z/x/y format
                    path = f"{tile[2]}/{tile[0]}/{tile[1]}"
                remote = url % path
            else:
                remote = url % filespec
            print("Getting file from: %s" % remote)
            # Create the subdirectories as pySmartDL doesn't do it for us
            Path(dest).mkdir(parents=True, exist_ok=True)

        dl = None
        try:
            if site["source"] == "topo":
                filespec += "." + site["suffix"]
            outfile = dest + "/" + filespec
            if not Path(outfile).exists():
                dl = SmartDL(remote, dest=outfile, connect_default_logger=False)
                dl.start()
            else:
                log.debug("%s exists!" % (outfile))
        except Exception as e:
            log.error(e)
            if dl:
                log.error(f"Couldn't download {filespec}: {dl.get_errors()}")
            else:
                log.error(f"Couldn't download {filespec}")


class BaseMapper(object):
    """Basemapper parent class."""

    def __init__(
        self,
        boundary: str,
        base: str,
        source: str,
        xy: bool,
    ):
        """Create an tile basemap for ODK Collect.

        Args:
            boundary (str): A BBOX string or GeoJSON file of the AOI.
                The GeoJSON can contain multiple geometries.
            base (str): The base directory to cache map tile in
            source (str): The upstream data source for map tiles
            xy (bool): Whether to swap the X & Y fields in the TMS URL

        Returns:
            (BaseMapper): An instance of this class
        """
        self.bbox = self.makeBbox(boundary)
        self.tiles = list()
        self.base = base
        # sources for imagery
        self.source = source
        self.sources = dict()
        self.xy = xy

        path = xlsforms_path.replace("xlsforms", "imagery.yaml")
        self.yaml = YamlFile(path)

        for entry in self.yaml.yaml["sources"]:
            for k, v in entry.items():
                src = dict()
                for item in v:
                    src["source"] = k
                    for k1, v1 in item.items():
                        # print(f"\tFIXME2: {k1} - {v1}")
                        src[k1] = v1
                self.sources[k] = src

    def customTMS(self, url: str, name: str = "custom", source: str = "custom", suffix: str = "jpg"):
        """Add a custom TMS URL to the list of sources.

        The url must end in %s to be replaced with the tile xyz values.

        Format examples:
        https://basemap.nationalmap.gov/ArcGIS/rest/services/USGSTopo/MapServer/tile/{z}/{y}/{x}
        https://maps.nyc.gov/xyz/1.0.0/carto/basemap/%s
        https://maps.nyc.gov/xyz/1.0.0/carto/basemap/{z}/{x}/{y}.jpg

        The method will replace {z}/{x}/{y}.jpg with %s

        Args:
            name (str): The name to display
            url (str): The URL string
            suffix (str): The suffix, png or jpg
            source (str): The source value to use as an index
        """
        # Remove any file extensions if present and update the 'suffix' parameter
        if url.endswith(".jpg"):
            source = "jpg"
            suffix = "jpg"
            url = url[:-4]  # Remove the last 4 characters (".jpg")
        elif url.endswith(".png"):
            source = "png"
            suffix = "png"
            url = url[:-4]  # Remove the last 4 characters (".png")

        # Replace "{z}/{x}/{y}" with "%s"
        url = re.sub(r"/{[xyz]+\}", "", url)
        url = url + r"/%s"

        tms_params = {"name": name, "url": url, "suffix": suffix, "source": source}
        log.debug(f"Setting custom TMS with params: {tms_params}")
        self.sources["custom"] = tms_params
        self.source = "custom"

    def getFormat(self):
        """Get the image format of the map tiles.

        Returns:
            (str): the upstream source for map tiles.
        """
        return self.sources[self.source]["suffix"]

    def getTiles(
        self,
        zoom: int = None,
    ):
        """Get a list of tiles for the specifed zoom level.

        Args:
            zoom (int): The Zoom level of the desired map tiles

        Returns:
            (int): The total number of map tiles downloaded
        """
        info = get_cpu_info()
        cores = info["count"]

        self.tiles = list(mercantile.tiles(self.bbox[0], self.bbox[1], self.bbox[2], self.bbox[3], zoom))
        total = len(self.tiles)
        log.info("%d tiles for zoom level %d" % (len(self.tiles), zoom))
        chunk = round(len(self.tiles) / cores)
        queue.Queue(maxsize=cores)
        log.info("%d threads, %d tiles" % (cores, total))

        mirrors = [self.sources[self.source]]
        # epdb.st()
        if len(self.tiles) < chunk or chunk == 0:
            dlthread(self.base, mirrors, self.tiles, self.xy)
        else:
            with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as executor:
                # results = []
                block = 0
                while block <= len(self.tiles):
                    executor.submit(dlthread, self.base, mirrors, self.tiles[block : block + chunk], self.xy)
                    # result = executor.submit(dlthread, self.base, mirrors, self.tiles[block : block + chunk], self.xy)
                    # results.append(result)
                    log.debug("Dispatching Block %d:%d" % (block, block + chunk))
                    block += chunk
                executor.shutdown()
            # log.info("Had %r errors downloading %d tiles for data for %r" % (self.errors, len(tiles), Path(self.base).name))
            # for result in results:
            #     print(result.result())
        return len(self.tiles)

    def tileExists(
        self,
        tile: MapTile,
    ):
        """See if a map tile already exists.

        Args:
            tile (MapTile): The map tile to check for the existence of

        Returns:
            (bool): Whether the tile exists in the map tile cache
        """
        filespec = f"{self.base}{tile[2]}/{tile[1]}/{tile[0]}.{self.sources[{self.source}]['suffix']}"
        if Path(filespec).exists():
            log.debug("%s exists" % filespec)
            return True
        else:
            log.debug("%s doesn't exists" % filespec)
            return False

    def makeBbox(
        self,
        boundary: str,
    ) -> tuple[float, float, float, float]:
        """Make a bounding box from a shapely geometry.

        Args:
            boundary (str): A BBOX string or GeoJSON file of the AOI.
                The GeoJSON can contain multiple geometries.

        Returns:
            (list): The bounding box coordinates
        """
        if not boundary.lower().endswith((".json", ".geojson")):
            # Is BBOX string
            try:
                if "," in boundary:
                    bbox_parts = boundary.split(",")
                else:
                    bbox_parts = boundary.split(" ")
                bbox = tuple(float(x) for x in bbox_parts)
                if len(bbox) == 4:
                    # BBOX valid
                    return bbox
                else:
                    msg = f"BBOX string malformed: {bbox}"
                    log.error(msg)
                    raise ValueError(msg) from None
            except Exception as e:
                log.error(e)
                msg = f"Failed to parse BBOX string: {boundary}"
                log.error(msg)
                raise ValueError(msg) from None

        log.debug(f"Reading geojson file: {boundary}")
        with open(boundary, "r") as f:
            poly = geojson.load(f)
        if "features" in poly:
            geometry = shape(poly["features"][0]["geometry"])
        elif "geometry" in poly:
            geometry = shape(poly["geometry"])
        else:
            geometry = shape(poly)

        if isinstance(geometry, list):
            # Multiple geometries
            log.debug("Creating union of multiple bbox geoms")
            geometry = unary_union(geometry)

        if geometry.is_empty:
            msg = f"No bbox extracted from {geometry}"
            log.error(msg)
            raise ValueError(msg) from None

        bbox = geometry.bounds
        # left, bottom, right, top
        # minX, minY, maxX, maxY
        return bbox


def tileid_from_y_tile(filepath: Union[Path | str]):
    """Helper function to get the tile id from a tile in z/x/y directory structure.

    Args:
        filepath (Union[Path, str]): The path to the y tile in /z/x/y.jpg structure.
    """
    # Get final two dirs + tile filename
    parts = list(Path(filepath).parts[-3:])
    # strip extension from y tile filename
    parts[-1] = str(Path(parts[-1]).stem)
    z, x, y = map(int, parts)
    return zxy_to_tileid(z, x, y)


def tile_dir_to_pmtiles(outfile: str, tile_dir: str, bbox: tuple, attribution: str):
    """Write PMTiles archive from tiles in the specified directory.

    Args:
        outfile (str): The output PMTiles archive file path.
        tile_dir (str): The directory containing the tile images.
        bbox (tuple): Bounding box in format (min_lon, min_lat, max_lon, max_lat).
        attribution (str): Attribution string to include in PMTile archive.

    Returns:
        None
    """
    tile_dir = Path(tile_dir)

    # Get tile image format from the first file encountered
    first_file = next((file for file in tile_dir.rglob("*.*") if file.is_file()), None)

    if not first_file:
        err = "No tile files found in the specified directory. Aborting PMTile creation."
        log.error(err)
        raise ValueError(err)

    # FIXME passing as PMTileType[tile_format] does not work
    # tile_format = first_file.suffix.upper()

    # Get zoom levels from dirs
    zoom_levels = sorted([int(x.stem) for x in tile_dir.glob("*") if x.is_dir()])

    # Process tiles
    with open(outfile, "wb") as pmtile_file:
        writer = PMTileWriter(pmtile_file)

        for tile_path in tile_dir.rglob("*"):
            if tile_path.is_file():
                tile_id = tileid_from_y_tile(tile_path)

                with open(tile_path, "rb") as tile:
                    writer.write_tile(tile_id, tile.read())

        # Extract bbox values
        min_lon, min_lat, max_lon, max_lat = bbox

        # Write metadata
        writer.finalize(
            {
                # "tile_type": TileType[tile_format.lstrip(".")],
                "tile_type": PMTileType.PNG,
                "tile_compression": PMTileCompression.NONE,
                "min_zoom": zoom_levels[0],
                "max_zoom": zoom_levels[-1],
                "min_lon_e7": int(min_lon * 10000000),
                "min_lat_e7": int(min_lat * 10000000),
                "max_lon_e7": int(max_lon * 10000000),
                "max_lat_e7": int(max_lat * 10000000),
                "center_zoom": zoom_levels[0],
                "center_lon_e7": int(min_lon + ((max_lon - min_lon) / 2)),
                "center_lat_e7": int(min_lat + ((max_lat - min_lat) / 2)),
            },
            {"attribution": f"© {attribution}"},
        )


def create_basemap_file(
    boundary=None,
    tms=None,
    xy=False,
    outfile=None,
    zooms="12-17",
    outdir=None,
    source="esri",
):
    """Create a basemap with given parameters.

    Args:
        boundary (str, optional): The boundary for the area you want.
        tms (str, optional): Custom TMS URL.
        xy (bool, optional): Swap the X & Y coordinates when using a
            custom TMS if True.
        outfile (str, optional): Output file name for the basemap.
        zooms (str, optional): The Zoom levels, specified as a range
            (e.g., "12-17") or comma-separated levels (e.g., "12,13,14").
        outdir (str, optional): Output directory name for tile cache.
        source (str, optional): Imagery source, one of
            ["esri", "bing", "topo", "google", "oam"] (default is "esri").

    Returns:
        None
    """
    log.debug(
        "Creating basemap with params: "
        f"boundary={boundary} | "
        f"outfile={outfile} | "
        f"zooms={zooms} | "
        f"outdir={outdir} | "
        f"source={source} | "
        f"xy={xy} | "
        f"tms={tms}"
    )

    # Validation
    if not boundary:
        err = "You need to specify a boundary! (file or bbox)"
        log.error(err)
        raise ValueError(err)

    # Get all the zoom levels we want
    zoom_levels = list()
    if zooms:
        if zooms.find("-") > 0:
            start = int(zooms.split("-")[0])
            end = int(zooms.split("-")[1]) + 1
            x = range(start, end)
            for i in x:
                zoom_levels.append(i)
        elif zooms.find(",") > 0:
            levels = zooms.split(",")
            for level in levels:
                zoom_levels.append(int(level))
        else:
            zoom_levels.append(int(zooms))

    if not outdir:
        base = Path.cwd().absolute()
    else:
        base = Path(outdir).absolute()

    source = "custom" if tms else source
    tiledir = base / f"{source}tiles"
    # Make tile download directory
    tiledir.mkdir(parents=True, exist_ok=True)
    # Convert to string for other methods
    tiledir = str(tiledir)

    if not source and not tms:
        err = "You need to specify a source!"
        log.error(err)
        raise ValueError(err)

    basemap = BaseMapper(boundary, tiledir, source, xy)

    if tms:
        # Add TMS URL to sources for download
        basemap.customTMS(tms)

    # Args parsed, main code:
    tiles = list()
    for level in zoom_levels:
        # Download the tile directory
        basemap.getTiles(level)
        tiles += basemap.tiles

    if not outfile:
        log.info(f"No outfile specified, tile download finished: {tiledir}")
        return

    suffix = Path(outfile).suffix.lower()
    log.debug(f"Basemap output format: {suffix}")

    if any(substring in suffix for substring in ["sqlite", "mbtiles"]):
        outf = DataFile(outfile, basemap.getFormat())
        if suffix == ".mbtiles":
            outf.addBounds(basemap.bbox)
        # Create output database and specify image format, png, jpg, or tif
        outf.writeTiles(tiles, tiledir)

    elif suffix == ".pmtiles":
        tile_dir_to_pmtiles(outfile, tiledir, basemap.bbox, source)

    else:
        msg = f"Format {suffix} not supported"
        log.error(msg)
        raise ValueError(msg) from None
    log.info(f"Wrote {outfile}")


def main():
    """This main function lets this class be run standalone by a bash script."""
    parser = argparse.ArgumentParser(description="Create an tile basemap for ODK Collect")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument(
        "-b",
        "--boundary",
        nargs="*",
        required=True,
        help=(
            "The boundary for the area you want. " "Accepts path to geojson file or bbox string. " "Format min_x min_y max_x max_y"
        ),
    )
    parser.add_argument("-t", "--tms", help="Custom TMS URL")
    parser.add_argument("--xy", default=False, help="Swap the X & Y coordinates when using a custom TMS")
    parser.add_argument(
        "-o", "--outfile", required=False, help="Output file name, allowed extensions [.mbtiles/.sqlitedb/.pmtiles]"
    )
    parser.add_argument("-z", "--zooms", default="12-17", help="The Zoom levels")
    parser.add_argument("-d", "--outdir", help="Output directory name for tile cache")
    parser.add_argument(
        "-s",
        "--source",
        default="esri",
        choices=["esri", "bing", "topo", "google", "oam"],
        help="Imagery source",
    )
    args = parser.parse_args()

    if not args.boundary:
        log.error("You need to specify a boundary! (file or bbox)")
        parser.print_help()
        quit()

    if not args.source:
        log.error("You need to specify a source!")
        parser.print_help()
        quit()

    if len(args.boundary) == 1:
        if Path(args.boundary[0]).suffix not in [".json", ".geojson"]:
            log.error("")
            log.error("*Error*: the boundary file must have .json or .geojson extension")
            log.error("")
            parser.print_help()
            quit()
        boundary_parsed = args.boundary[0]
    elif len(args.boundary) == 4:
        boundary_parsed = ",".join(args.boundary)
    else:
        log.error("")
        log.error("*Error*: the bounding box must have 4 coordinates")
        log.error("")
        parser.print_help()
        quit()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    create_basemap_file(
        boundary=boundary_parsed,
        tms=args.tms,
        xy=args.xy,
        outfile=args.outfile,
        zooms=args.zooms,
        outdir=args.outdir,
        source=args.source,
    )


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import logging
import sys

from osm_fieldwork.xlsforms import xlsforms_path
from osm_fieldwork.yamlfile import YamlFile

# Instantiate logger
log = logging.getLogger(__name__)


def escape(value: str):
    """Escape characters like embedded quotes in text fields.

    Args:
        value (str):The string to modify

    Returns:
        (str): The escaped string
    """
    # tmp = value.replace(" ", "_")
    tmp = value.replace("&", " and ")
    return tmp.replace("'", "&apos;")


class Convert(YamlFile):
    """A class to apply a YAML config file and convert ODK to OSM.

    Returns:
        (Convert): An instance of this object
    """

    def __init__(
        self,
        xform: str = None,
    ):
        path = xlsforms_path.replace("xlsforms", "")
        if xform is not None:
            file = xform
        else:
            file = f"{path}/xforms.yaml"
        self.yaml = YamlFile(file)
        self.filespec = file
        # Parse the file contents into a data structure to make it
        # easier to retrieve values
        self.convert = dict()
        self.ignore = list()
        self.private = list()
        for item in self.yaml.yaml["convert"]:
            key = list(item.keys())[0]
            value = item[key]
            # print("ZZZZ: %r, %r" % (key, value))
            if type(value) is str:
                self.convert[key] = value
            elif type(value) is list:
                vals = dict()
                for entry in value:
                    if type(entry) is str:
                        # epdb.st()
                        tag = entry
                    else:
                        tag = list(entry.keys())[0]
                        vals[tag] = entry[tag]
                self.convert[key] = vals
        self.ignore = self.yaml.yaml["ignore"]
        self.private = self.yaml.yaml["private"]
        if "multiple" in self.yaml.yaml:
            self.multiple = self.yaml.yaml["multiple"]
        else:
            self.multiple = list()

    def privateData(
        self,
        keyword: str,
    ):
        """See is a keyword is in the private data category.

        Args:
            keyword (str): The keyword to search for

        Returns:
            (bool): Check to see if the keyword is in the private data section
        """
        return keyword.lower() in self.private

    def convertData(
        self,
        keyword: str,
    ):
        """See is a keyword is in the convert data category.

        Args:
            keyword (str): The keyword to search for

        Returns:
            (bool): Check to see if the keyword is in the convert data section
        """
        return keyword.lower() in self.convert

    def ignoreData(
        self,
        keyword: str,
    ):
        """See is a keyword is in the convert data category.

        Args:
            keyword (str): The keyword to search for

        Returns:
            (bool): Check to see if the keyword is in the ignore data section
        """
        return keyword.lower() in self.ignore

    def getKeyword(
        self,
        value: str,
    ):
        """Get the keyword for a value from the yaml file.

        Args:
            value (str): The value to find the keyword for
        Returns:
            (str): The keyword if found, or None
        """
        key = self.yaml.yaml(value)
        if type(key) == bool:
            return value
        if len(key) == 0:
            key = self.yaml.getKeyword(value)
        return key

    def getValues(
        self,
        keyword: str = None,
    ):
        """Get the values for a primary key.

        Args:
            keyword (str): The keyword to get the value of

        Returns:
            (str): The values or None
        """
        if keyword is not None:
            if keyword in self.convert:
                return self.convert[keyword]
        else:
            return None

    def convertEntry(
        self,
        tag: str,
        value: str,
    ):
        """Convert a tag and value from the ODK represention to an OSM one.

        Args:
            tag (str): The tag from the ODK XML file
            value (str): The value from the ODK XML file

        Returns:
            (list): The converted values
        """
        all = list()

        # If it's not in any conversion data, pass it through unchanged.
        if tag.lower() in self.ignore:
            # logging.debug(f"FIXME: Ignoring {tag}")
            return None
        low = tag.lower()
        if low not in self.convert and low not in self.ignore and low not in self.private:
            return {tag: value}

        newtag = tag.lower()
        newval = value
        # If the tag is in the config file, convert it.
        if self.convertData(newtag):
            newtag = self.convertTag(newtag)
            if newtag != tag:
                logging.debug(f"Converted Tag for entry {tag} to {newtag}")

        # Truncate the elevation, as it's really long
        if newtag == "ele":
            value = value[:7]
        newval = self.convertValue(newtag, value)
        logging.debug("Converted Value for entry '%s' to '%s'" % (value, newval))
        # there can be multiple new tag/value pairs for some values from ODK
        if type(newval) == str:
            all.append({newtag: newval})
        elif type(newval) == list:
            for entry in newval:
                if type(entry) == str:
                    all.append({newtag: newval})
                elif type(entry) == dict:
                    for k, v in entry.items():
                        all.append({k: v})
        return all

    def convertValue(
        self,
        tag: str,
        value: str,
    ):
        """Convert a single tag value.

        Args:
            tag (str): The tag from the ODK XML file
            value (str): The value from the ODK XML file

        Returns:
            (list): The converted values
        """
        all = list()

        vals = self.getValues(tag)
        # There is no conversion data for this tag
        if vals is None:
            return value

        if type(vals) is dict:
            if value not in vals:
                all.append({tag: value})
                return all
            if type(vals[value]) is bool:
                entry = dict()
                if vals[value]:
                    entry[tag] = "yes"
                else:
                    entry[tag] = "no"
                all.append(entry)
                return all
            for item in vals[value].split(","):
                entry = dict()
                tmp = item.split("=")
                if len(tmp) == 1:
                    entry[tag] = vals[value]
                else:
                    entry[tmp[0]] = tmp[1]
                    logging.debug("\tValue %s converted to %s" % (value, entry))
                all.append(entry)
        return all

    def convertTag(
        self,
        tag: str,
    ):
        """Convert a single tag.

        Args:
            tag (str): The tag from the ODK XML file

        Returns:
            (str): The new tag
        """
        low = tag.lower()
        if low in self.convert:
            newtag = self.convert[low]
            if type(newtag) is str:
                logging.debug("\tTag '%s' converted to '%s'" % (tag, newtag))
                tmp = newtag.split("=")
                if len(tmp) > 1:
                    newtag = tmp[0]
            elif type(newtag) is list:
                logging.error("FIXME: list()")
                # epdb.st()
                return low
            elif type(newtag) is dict:
                # logging.error("FIXME: dict()")
                return low
            return newtag.lower()
        else:
            return low

    def dump(self):
        """Dump internal data structures, for debugging purposes only."""
        print("YAML file: %s" % self.filespec)
        print("Convert section")
        for key, val in self.convert.items():
            if type(val) is list:
                print("\tTag %s is" % key)
                for data in val:
                    print("\t\t%r" % data)
            else:
                print("\tTag %s is %s" % (key, val))

        print("Ignore Section")
        for item in self.ignore:
            print(f"\tIgnoring tag {item}")


#
# This script can be run standalone for debugging purposes. It's easier to debug
# this way than using pytest,
#
def main():
    """This main function lets this class be run standalone by a bash script."""
    parser = argparse.ArgumentParser(description="Read and parse a YAML file")

    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-x", "--xform", default="xform.yaml", help="Default Yaml file")
    parser.add_argument(
        "-i",
        "--infile",
        required=True,
        help="The CSV input file",
    )
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    # convert = Convert(args.xform)
    convert = Convert("xforms.yaml")
    print("-----")
    # tag = convert.convertTag("waterpoint_seasonal")
    # entry = convert.convertEntry("waterpoint_seasonal")
    # print("YY: %r" % entry)

    # print(convert.convertTag("tourism"))
    # entry = convert.convertEntry("tourism")
    # print(entry)
    # value = convert.convertEntry("waterpoint_seasonal")
    # print(value)
    # print("===============")

    # tag = convert.convertTag("waterpoint")
    # print(tag)
    # value = convert.convertValue("waterpoint", "well")
    # print(value)
    # value = convert.convertValue("power", "solar")

    entry = convert.convertEntry("waterpoint", "faucet")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("operational_status", "closed")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("seasonal", "wet")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("seasonal", "rainy")
    for i in entry:
        print("XX: %r" % i)

    entry = convert.convertEntry("power", "solar")
    for i in entry:
        print("XX: %r" % i)


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of Osm-Fieldwork.
#
#     Osm-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Osm-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import logging
import sqlite3
import sys

import pandas as pd
import requests

#
# This program is a utility to validate tags & values from the
# Impact spreadsheet with what is defined on the OSM Wiki
# or in taginfo, which is all tags every used in OSM. It's purpose
# is to create the private data section for ODK->OSM conversion.
#

# Instantiate logger
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class ValidateModel(object):
    def __init__(self, taginfo=None):
        self.tags = dict()
        if not taginfo:
            self.taginfo = "taginfo-db.db"
        else:
            self.taginfo = taginfo
        self.db = sqlite3.connect(self.taginfo)
        self.cursor = self.db.cursor()
        self.threshold = dict()
        self.url = "https://taginfo.openstreetmap.org/taginfo/apidoc#api_4_key_values?"
        self.session = requests.Session()
        self.headers = {"Content-Type": "application/json;charset=UTF-8"}
        self.threshold = 100

    def parse(self):
        models = "Impact Areas - Data Models V1.1.xlsx"
        data = pd.read_excel(models, sheet_name="Overview - all Tags", usecols=["key", "value"])

        entries = data.to_dict()
        total = len(entries["key"])
        index = 1
        while index < total:
            key = entries["key"][index]
            value = entries["value"][index]
            if value == "<text>":
                index += 1
                continue
            if key not in self.tags:
                self.tags[key.strip()] = list()
            self.tags[key.strip()].append(value.strip())
            index += 1
        return self.tags

    def validateTaginfo(self, csv=None):
        if csv:
            csvfile = open(csv, "w")
        threshold = self.threshold
        for key in self.tags.keys():
            for value in self.tags[key]:
                if value[:3] == "yes" or value[:2] == "no" or value[0] == "<":
                    continue
                threshold = self.threshold
                sql = f"SELECT value,count_all FROM tags where key='{key}'"
                # logging.debug(sql)
                result = self.cursor.execute(sql)
                data = result.fetchall()
                if len(data) == 0:
                    logging.warning(f"'{key}' does not exist in taginfo!")
                else:
                    for val in data:
                        if val[0] == value:
                            if value[:3] == "yes" or value[:2] == "no" or value[0] == "<":
                                continue
                            # logging.info(f"\"{value}\" exists in the taginfo for \"{key}\"!")
                        if val[1] < threshold:
                            logging.warning(f'"{value}" doesn\'t pass the threshold for "{key}"! Only {val[1]} occurances')
                            if csv:
                                csvfile.write(f"{key},{value},{val[1]}\n")
                            break


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Validate data_models using taginfo database")
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument("-t", "--taginfo", help="Taginfo database")
    parser.add_argument("-c", "--csv", help="Output CSV")
    args = parser.parse_args()

    # if verbose, dump to the termina
    if not args.verbose:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch.setFormatter(formatter)
        root.addHandler(ch)

    model = ValidateModel(args.taginfo)
    tags = model.parse()
    # import epdb; epdb.st()
    model.validateTaginfo(args.csv)



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import concurrent.futures
import logging
import os
import sys

import geojson
import shapely
from codetiming import Timer
from cpuinfo import get_cpu_info
from geojson import Point
from haversine import Unit, haversine
from osm_rawdata.postgres import PostgresClient
from shapely.geometry import mapping, shape
from thefuzz import fuzz

from osm_fieldwork.convert import escape
from osm_fieldwork.data_models import data_models_path
from osm_fieldwork.osmfile import OsmFile

# Instantiate logger
log = logging.getLogger(__name__)

# The number of threads is based on the CPU cores
info = get_cpu_info()
cores = info["count"]


class OdkMerge(object):
    def __init__(
        self,
        source: str,
        boundary: str = None,
    ):
        """Initialize Input data source.

        Args:
            source (str):
            boundary: str = None

        Returns:
            (OdkMerge): An instance of this object
        """
        self.postgres = list()
        self.source = source
        self.tags = dict()
        # Distance in meters for conflating with postgis
        self.tolerance = 7
        self.data = dict()
        self.analyze = ("name", "amenity", "landuse", "cuisine", "tourism", "leisure")
        # PG: is the same prefix as ogr2ogr
        # "[user[:password]@][netloc][:port][/dbname]"
        if source[0:3] == "PG:":
            uri = source[3:]
            # self.source = "underpass" is not support yet
            # Each thread needs it's own connection to postgres to avoid problems.
            for _thread in range(0, cores + 1):
                db = PostgresClient(uri, f"{data_models_path}/{config}")
                self.postgres.append(db)
                if boundary:
                    self.clip(boundary, db)
        else:
            log.info("Opening data file: %s" % source)
            src = open(source, "r")
            self.data = geojson.load(src)
            if boundary:
                self.clip(boundary)

    def clip(
        self,
        boundary: str,
        db: PostgresClient,
    ):
        """Clip a data source by a boundary.

        Args:
             boundary (str): The filespec of the project AOI
             db (PostgresClient): A reference to the existing database connection

        Returns:
            (bool): If the region was clipped sucessfully
        """
        if not boundary:
            return False

        if type(boundary) != dict:
            clip = open(boundary, "r")
            geom = geojson.load(clip)
            if "features" in geom:
                poly = geom["features"][0]["geometry"]
            else:
                poly = geom["geometry"]
        else:
            poly = boundary
        ewkt = shape(poly)

        if len(self.data) > 0:
            # TODO: FMTM produces data extracts the exact size of the boundary
            # polygon, so  we don't need to clip it. In the future though we
            # want this to produce a subset from a larger file.
            for feature in self.data["features"]:
                shapely.from_geojson(str(feature))
            #     if not shapely.contains(ewkt, entry):
            #         log.debug(f"CONTAINS {entry}")
            #         del self.data[self.data['features']]
            pass
        else:
            # setup the postgres VIEWs with a dummy SQL query
            sql = "SELECT COUNT(osm_id) FROM nodes"
            db.queryLocal(sql, ewkt)
        return True

    def makeNewFeature(
        self,
        attrs: dict = None,
        tags: dict = None,
    ):
        """Create a new feature with optional data.

        Args:
            attrs (dict): All of the attributes and their values
            tags (dict): All of the tags and their values

        Returns:
            (dict): A template feature with no data
        """
        newf = dict()
        if attrs:
            newf["attrs"] = attrs
        else:
            newf["attrs"] = dict()
        if tags:
            newf["tags"] = tags
        else:
            newf["tags"] = dict()
        return newf

    def conflateFile(
        self,
        feature: dict,
    ):
        """Conflate a POI against all the features in a GeoJson file.

        Args:
            feature (dict): The feature to conflate

        Returns:
            (dict):  The modified feature
        """
        # Most smartphone GPS are 5-10m off most of the time, plus sometimes
        # we're standing in front of an amenity and recording that location
        # instead of in the building.
        gps_accuracy = 10
        # this is the treshold for fuzzy string matching
        match_threshold = 80
        # log.debug(f"conflateFile({feature})")
        hits = False
        geom = Point((float(feature["attrs"]["lon"]), float(feature["attrs"]["lat"])))
        wkt = shape(geom)
        for existing in self.data["features"]:
            id = int(existing["properties"]["id"])
            entry = shapely.from_geojson(str(existing))
            if entry.geom_type != "Point":
                center = shapely.centroid(entry)
            else:
                center = entry
                # dist = shapely.hausdorff_distance(center, wkt)
                # if 'name' in existing['properties']:
                #     print(f"DIST1: {dist}, {existing['properties']['name']}")
            # x = shapely.distance(wkt, entry)
            # haversine reverses the order of lat & lon from what shapely uses. We
            # use this as meters is easier to deal with than cartesian coordinates.
            x1 = (center.coords[0][1], center.coords[0][0])
            x2 = (wkt.coords[0][1], wkt.coords[0][0])
            dist = haversine(x1, x2, unit=Unit.METERS)
            if dist < gps_accuracy:
                # if 'name' in existing['properties']:
                # log.debug(f"DIST2: {dist}")
                # log.debug(f"Got a Hit! {feature['tags']['name']}")
                for key, value in feature["tags"].items():
                    if key in self.analyze:
                        if key in existing["properties"]:
                            result = fuzz.ratio(value, existing["properties"][key])
                            if result > match_threshold:
                                # log.debug(f"Matched: {result}: {feature['tags']['name']}")
                                existing["properties"]["fixme"] = "Probably a duplicate!"
                                log.debug(f"Got a dup in file!!! {existing['properties']['name'] }")
                                hits = True
                                break
            if hits:
                version = int(existing["properties"]["version"])
                # coords = feature['geometry']['coordinates']
                # lat = coords[1]
                # lon = coords[0]
                attrs = {"id": id, "version": version, "lat": feature["attrs"]["lat"], "lon": feature["attrs"]["lon"]}
                tags = existing["properties"]
                tags["fixme"] = "Probably a duplicate!"
                # Data extracts for ODK Collect
                del tags["title"]
                del tags["label"]
                if "building" in tags:
                    return {"attrs": attrs, "tags": tags, "refs": list()}
                return {"attrs": attrs, "tags": tags}
        return dict()

    def conflateWay(
        self,
        feature: dict,
        dbindex: int,
    ):
        """Conflate a POI against all the ways in a postgres view.

        Args:
            feature (dict): The feature to conflate
            dbindex (int): An index into the array of postgres connections

        Returns:
            (dict):  The modified feature
        """
        # log.debug(f"conflateWay({feature})")
        hits = False
        result = list()
        geom = Point((float(feature["attrs"]["lon"]), float(feature["attrs"]["lat"])))
        wkt = shape(geom)
        for key, value in feature["tags"].items():
            if key in self.analyze:
                # Sometimes the duplicate is a polygon, really common for parking lots.
                cleanval = escape(value)
                query = f"SELECT osm_id,tags,version,ST_AsText(ST_Centroid(geom)) FROM ways_view WHERE ST_Distance(geom::geography, ST_GeogFromText('SRID=4326;{wkt.wkt}')) < {self.tolerance} AND levenshtein(tags->>'{key}', '{cleanval}') <= 1"
                # log.debug(query)
                self.postgres[dbindex].dbcursor.execute(query)
                try:
                    result = self.postgres[dbindex].dbcursor.fetchall()
                except:
                    result = list()
                    # log.warning(f"No results at all for {query}")
                if len(result) > 0:
                    hits = True
                    break
        if hits:
            log.debug(f"Got a dup in ways!!! {feature['tags']['name']}")
            # the result is a list from what we specify for SELECT
            version = int(result[0][2]) + 1
            attrs = {"id": int(result[0][0]), "version": version}
            tags = result[0][1]
            tags[f"old_{key}"] = value
            tags["fixme"] = "Probably a duplicate!"
            geom = mapping(shapely.from_wkt(result[0][3]))
            refs = list()
            # FIXME: iterate through the points and find the existing nodes,
            # which I'm not sure
            # is possible
            # SELECT osm_id,tags,version FROM nodes WHERE ST_Contains(geom, ST_GeomFromText('Point(-105.9918636 38.5360821)'));
            # for i in geom['coordinates'][0]:
            #    print(f"XXXXX: {i}")
            return {"attrs": attrs, "tags": tags, "refs": refs}
        return dict()

    def conflateNode(
        self,
        feature: dict,
        dbindex: int,
    ):
        """Conflate a POI against all the nodes in the view.

        Args:
            feature (dict): The feature to conflate
            dbindex (int): An index into the array of postgres connections

        Returns:
            (dict):  The modified feature
        """
        # log.debug(f"conflateNode({feature})")
        hits = False
        geom = Point((float(feature["attrs"]["lon"]), float(feature["attrs"]["lat"])))
        wkt = shape(geom)
        result = list()
        ratio = 1
        for key, value in feature["tags"].items():
            if key in self.analyze:
                # print("%s = %s" % (key, value))
                # Use a Geography data type to get the answer in meters, which
                # is easier to deal with than degress of the earth.
                cleanval = escape(value)
                query = f"SELECT osm_id,tags,version,ST_AsEWKT(geom) FROM nodes_view WHERE ST_Distance(geom::geography, ST_GeogFromText('SRID=4326;{wkt.wkt}')) < {self.tolerance} AND levenshtein(tags->>'{key}', '{cleanval}') <= {ratio}"
                # print(query)
                # FIXME: this currently only works with a local database, not underpass yet
                self.postgres[dbindex].dbcursor.execute(query)
                try:
                    result = self.postgres[dbindex].dbcursor.fetchall()
                except:
                    result = list()
                    # log.warning(f"No results at all for {query}")
                if len(result) > 0:
                    hits = True
                    break
        if hits:
            log.debug(f"Got a dup in nodes!!! {feature['tags']}")
            version = int(result[0][2]) + 1
            coords = shapely.from_wkt(result[0][3][10:])
            lat = coords.y
            lon = coords.x
            attrs = {"id": int(result[0][0]), "version": version, "lat": lat, "lon": lon}
            tags = result[0][1]
            tags[f"old_{key}"] = value
            tags["fixme"] = "Probably a duplicate!"
            return {"attrs": attrs, "tags": tags}
        return dict()

    def conflateById(
        self,
        feature: dict,
        dbindex: int,
    ):
        """Conflate a feature with existing ways using the OSM ID.

        Args:
            feature (dict): The feature to conflate
            dbindex (int): An index into the array of postgres connections

        Returns:
            (dict):  The modified feature
        """
        log.debug(f"conflateById({feature})")
        id = int(feature["attrs"]["id"])
        if id > 0:
            if self.source[:3] != "PG:":
                sql = f"SELECT osm_id,tags,version,ST_AsText(geom) FROM ways_view WHERE tags->>'id'='{id}'"
                # log.debug(sql)
                self.postgres[0].dbcursor.execute(sql)
                result = self.postgres[0].dbcursor.fetchone()
                if result:
                    version = int(result[0][2]) + 1
                    attrs = {"id": int(result[0][0]), "version": version}
                    tags = result[0][1]
                    # tags[f'old_{key}'] = value
                    tags["fixme"] = "Probably a duplicate!"
                    mapping(shapely.from_wkt(result[0][3]))
                    return {"attrs": attrs, "tags": tags}
                else:
                    sql = f"SELECT osm_id,tags,version,ST_AsText(geom) FROM ways_view WHERE tags->>'id'='{id}'"
                    # log.debug(sql)
                    self.postgres[dbindex].dbcursor.execute(sql)
                    result = self.postgres[dbindex].dbcursor.fetchone()
                    if result:
                        version = int(result[0][2]) + 1
                        attrs = {"id": int(result[0][0]), "version": version}
                        tags = result[0][1]
                        # tags[f'old_{key}'] = value
                        tags["fixme"] = "Probably a duplicate!"
                        mapping(shapely.from_wkt(result[0][3]))
                    return {"attrs": attrs, "tags": tags, "refs": refs}
            else:
                for key, value in self.data.items():
                    if key == id:
                        return value
        return dict()

    def cleanFeature(
        self,
        feature: dict,
    ):
        """Remove tags that are attributes instead
        Args:
            feature (dict): The feature to clean.

        Returns:
            (dict):  The modified feature
        """
        # We only use the version and ID in the attributes
        if "id" in feature["tags"]:
            del feature["tags"]["id"]
        if "version" in feature["tags"]:
            del feature["tags"]["version"]
        if "title" in feature["tags"]:
            del feature["tags"]["title"]
        if "label" in feature["tags"]:
            del feature["tags"]["label"]
        return feature

    def dump(self):
        """Dump internal data."""
        print(f"Data source is: {self.source}")
        print(f"There are {len(self.data)} existing features")
        # if len(self.versions) > 0:
        #     for k, v in self.original.items():
        #         print(f"{k}(v{self.versions[k]}) = {v}")

    def conflateData(
        self,
        odkdata: list,
    ):
        """Conflate all the data. This the primary interfacte for conflation.

        Args:
            odkdata (list): A list of all the entries in the OSM XML input file

        Returns:
            (dict):  The modified features
        """
        timer = Timer(text="conflateData() took {seconds:.0f}s")
        timer.start()
        # Use fuzzy string matching to handle minor issues in the name column,
        # which is often used to match an amenity.
        if len(self.data) == 0:
            self.postgres[0].dbcursor.execute("CREATE EXTENSION IF NOT EXISTS fuzzystrmatch")
        log.debug(f"OdkMerge::conflateData() called! {len(odkdata)} features")

        # A chunk is a group of threads
        chunk = round(len(odkdata) / cores)

        # cycle = range(0, len(odkdata), chunk)

        # Chop the data into a subset for each thread
        newdata = list()
        future = None
        result = None
        index = 0
        with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as executor:
            i = 0
            subset = dict()
            futures = list()
            for key, value in odkdata.items():
                subset[key] = value
                if i == chunk:
                    i = 0
                    result = executor.submit(conflateThread, subset, self, index)
                    index += 1
                    # result.add_done_callback(callback)
                    futures.append(result)
                    subset = dict()
                i += 1
            for future in concurrent.futures.as_completed(futures):
                # # for future in concurrent.futures.wait(futures, return_when='ALL_COMPLETED'):
                log.debug("Waiting for thread to complete..")
                # print(f"YYEESS!! {future.result(timeout=10)}")
                newdata.append(future.result(timeout=5))
        timer.stop()
        return newdata
        # return alldata


def conflateThread(
    features: dict,
    source: str,
    dbindex: int,
):
    """Conflate a subset of the data.

    Args:
        feature (dict): The feature to conflate
        source (str): The data source for conflation, file or database
        dbindex (int): An index into the array of postgres connections

    Returns:
    """
    timer = Timer(text="conflateThread() took {seconds:.0f}s")
    timer.start()
    log.debug(f"conflateThread() called! {len(features)} features")
    merged = list()
    result = dict()
    dups = 0
    # This is brute force, slow but accurate. Process each feature
    # and look for a possible match with existing data.
    for _key, value in features.items():
        id = int(value["attrs"]["id"])
        # Each of the conflation methods take a single feature
        # as a parameter, and returns a possible match or a zero
        # length dictionary.
        if id > 0:
            # Any feature ID greater than zero is existing data.
            if source.source[:3] != "PG:":
                result = source.conflateFile(value)
            else:
                # Any feature ID less than zero is new data collected
                # using geopoint in the XLSForm.
                result = source.conflateById(value, dbindex)
        elif id < 0:
            if source.source[:3] != "PG:":
                result = source.conflateFile(value)
            else:
                result = source.conflateNode(value, dbindex)
                if len(result) == 0:
                    result = source.conflateWay(value, dbindex)
        if result and len(result) > 0:
            # Merge the tags and attributes together, the OSM data and ODK data.
            # If no match is found, the ODK data is used to create a new feature.
            if "fixme" in result["tags"]:
                dups += 1
                # newf = source.cleanFeature(result)
                attrs = value["attrs"] | result["attrs"]
                tags = value["tags"] | result["tags"]
                merged.append({"attrs": attrs, "tags": tags})
            else:
                merged.append(value)
        else:
            merged.append(value)
            # log.error(f"There are no results!")
    timer.stop()
    log.debug(f"Found {dups} duplicates")
    return merged


def main():
    """This main function lets this class be run standalone by a bash script."""
    parser = argparse.ArgumentParser(
        prog="odk_merge.py",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="This program conflates ODK data with existing features from OSM.",
        epilog="""
    This program conflates the output from Osm-Fieldwork after converting
    the JSON file from ODK Central to OSM XML. The data source for existing data can'
    be either the data extract used by the XLSForm, or a postgresql database.

    examples:
        odk_merge.py [OPTIONS] [OSMFILE] [DATAFILE]
        Where OSMFILE is the output from json2osm.py
        And DATAFILE is either the data extract from make_data_extract.py,
        OR refix with "PG:[url]" to use a postgres database
        """,
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-o", "--outfile", help="Output file from the conflation")
    # parser.add_argument("-e", "--extract",  help="OSM data extract created by Osm-Fieldwork")
    # parser.add_argument("-f", "--odkfile",  required=True, help="OSM XML file created by Osm-Field")
    parser.add_argument("-b", "--boundary", help="Boundary polygon to limit the data size")

    args, unknown = parser.parse_known_args()
    osmdata = None
    source = None
    if len(unknown) < 2:
        parser.print_help()
        quit()
    else:
        osmdata = unknown[0]
        source = unknown[1]

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    if args.outfile:
        outfile = args.outfile
    else:
        outfile = os.path.basename(osmdata.replace(".osm", "-foo.osm"))

    # This is the existing OSM data, a database or a file
    if args.boundary:
        extract = OdkMerge(source, args.boundary)
    else:
        extract = OdkMerge(source)

    if extract:
        odkf = OsmFile(outfile)  # output file
        osm = odkf.loadFile(osmdata)  # input file
        # odkf.dump()
    else:
        log.error("No ODK data source specified!")
        parser.print_help()
        quit()

    # This returns a list of lists of dictionaries. Each thread returns
    # a list of the features, and len(data) is thre number of CPU cores.
    data = extract.conflateData(osm)
    out = list()
    # print(data)
    for entry in data:
        # if 'refs' in feature or 'building' in feature['tags']:
        for feature in entry:
            if "refs" in feature:
                feature["refs"] = list()
                out.append(odkf.createWay(feature, True))
            else:
                out.append(odkf.createNode(feature, True))

    # out = ""
    # for id, feature in osm.items():
    #     result = extract.conflateFile(feature)
    #     if len(result) > 0:
    #         node = odkf.featureToNode(result)
    #     else:
    #         node = feature
    #     out += odkf.createNode(node, True)
    odkf.write(out)
    log.info(f"Wrote {outfile}")


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import csv
import logging
import os
import re
import sys

import pandas as pd
from geojson import Feature, FeatureCollection, Point, dump

from osm_fieldwork.convert import Convert
from osm_fieldwork.osmfile import OsmFile
from osm_fieldwork.xlsforms import xlsforms_path

# Instantiate logger
log = logging.getLogger(__name__)


class CSVDump(Convert):
    """A class to parse the CSV files from ODK Central."""

    def __init__(
        self,
        yaml: str = None,
    ):
        self.fields = dict()
        self.nodesets = dict()
        self.data = list()
        self.osm = None
        self.json = None
        self.features = list()
        xlsforms_path.replace("xlsforms", "")
        if yaml:
            pass
        else:
            pass
        self.config = super().__init__(yaml)
        self.saved = dict()
        self.defaults = dict()

    def lastSaved(
        self,
        keyword: str,
    ):
        if keyword is not None and len(keyword) > 0:
            return self.saved[keyword]
        return None

    def updateSaved(
        self,
        keyword: str,
        value: str,
    ):
        if keyword is not None and value is not None and len(value) > 0:
            self.saved[keyword] = value

    def parseXLS(
        self,
        xlsfile: str,
    ):
        """Parse the source XLSFile if available to look for details we need."""
        if xlsfile is not None and len(xlsfile) > 0:
            entries = pd.read_excel(xlsfile, sheet_name=[0])
            # There will only be a single sheet
            names = entries[0]["name"]
            defaults = entries[0]["default"]
            total = len(names)
            i = 0
            while i < total:
                entry = defaults[i]
                if str(entry) != "nan":
                    pat = re.compile("..last-saved.*")
                    if pat.match(entry):
                        name = entry.split("#")[1][:-1]
                        self.saved[name] = None
                    else:
                        self.defaults[names[i]] = entry
                i += 1
        return True

    def createOSM(
        self,
        filespec: str,
    ):
        """Create an OSM XML output files."""
        log.debug("Creating OSM XML file: %s" % filespec)
        self.osm = OsmFile(filespec)
        # self.osm.header()

    def writeOSM(
        self,
        feature: dict,
    ):
        """Write a feature to an OSM XML output file."""
        out = ""
        if "id" in feature["tags"]:
            feature["id"] = feature["tags"]["id"]
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        if "refs" not in feature:
            out += self.osm.createNode(feature)
        else:
            out += self.osm.createWay(feature)
        self.osm.write(out)

    def finishOSM(self):
        """Write the OSM XML file footer and close it."""
        self.osm.footer()

    def createGeoJson(
        self,
        file: str = "tmp.geojson",
    ):
        """Create a GeoJson output file."""
        log.debug("Creating GeoJson file: %s" % file)
        self.json = open(file, "w")

    def writeGeoJson(
        self,
        feature: dict,
    ):
        """Write a feature to a GeoJson output file."""
        # These get written later when finishing , since we have to create a FeatureCollection
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        self.features.append(feature)

    def finishGeoJson(self):
        """Write the GeoJson FeatureCollection to the output file and close it."""
        features = list()
        for item in self.features:
            poi = Point((float(item["attrs"]["lon"]), float(item["attrs"]["lat"])))
            if "private" in item:
                props = {**item["tags"], **item["private"]}
            else:
                props = item["tags"]
            features.append(Feature(geometry=poi, properties=props))
        collection = FeatureCollection(features)
        dump(collection, self.json)

    def parse(
        self,
        filespec: str,
        data: str = None,
    ):
        """Parse the CSV file from ODK Central and convert it to a data structure."""
        all_tags = list()
        if not data:
            f = open(filespec, newline="")
            reader = csv.DictReader(f, delimiter=",")
        else:
            reader = csv.DictReader(data, delimiter=",")
        for row in reader:
            tags = dict()
            # log.info(f"ROW: {row}")
            for keyword, value in row.items():
                if keyword is None or len(keyword) == 0:
                    continue

                base = self.basename(keyword).lower()
                # There's many extraneous fields in the input file which we don't need.
                if base is None or base in self.ignore or value is None:
                    continue
                # if base in self.multiple:
                #     epdb.st()
                #     entry = reader[keyword]
                #     for key, val in entry.items():
                #         print(key)75.66.108.181
                #         if key == "name":
                #             tags['name'] = val
                #     continue
                else:
                    # When using geopoint warmup, once the display changes to the map
                    # location, there is not always a value if the accuracy is way
                    # off. In this case use the warmup value, which is where we are
                    # standing anyway.
                    if base == "latitude" and len(value) == 0:
                        if "warmup-Latitude" in row:
                            value = row["warmup-Latitude"]
                            if base == "longitude" and len(value) == 0:
                                value = row["warmup-Longitude"]
                    items = self.convertEntry(base, value)
                    # log.info(f"ROW: {base} {value}")
                    if len(items) > 0:
                        if base in self.saved:
                            if str(value) == "nan" or len(value) == 0:
                                # log.debug(f"FIXME: {base} {value}")
                                val = self.saved[base]
                                if val and len(value) == 0:
                                    log.warning(f'Using last saved value for "{base}"! Now "{val}"')
                                    value = val
                            else:
                                self.saved[base] = value
                                log.debug(f'Updating last saved value for "{base}" with "{value}"')
                        # Handle nested dict in list
                        if isinstance(items, list):
                            items = items[0]
                        for k, v in items.items():
                            tags[k] = v
                    else:
                        tags[base] = value
                # log.debug(f"\tFIXME1: {tags}")
            all_tags.append(tags)
        return all_tags

    def basename(
        self,
        line: str,
    ):
        """Extract the basename of a path after the last -."""
        tmp = line.split("-")
        if len(tmp) == 0:
            return line
        base = tmp[len(tmp) - 1]
        return base

    def createEntry(
        self,
        entry: dict,
    ):
        """Create the feature data structure."""
        # print(line)
        feature = dict()
        attrs = dict()
        tags = dict()
        priv = dict()
        refs = list()

        # log.debug("Creating entry")
        # First convert the tag to the approved OSM equivalent
        if "lat" in entry and "lon" in entry:
            attrs["lat"] = entry["lat"]
            attrs["lon"] = entry["lon"]
        for key, value in entry.items():
            attributes = (
                "id",
                "timestamp",
                "lat",
                "lon",
                "uid",
                "user",
                "version",
                "action",
            )

            # When using existing OSM data, there's a special geometry field.
            # Otherwise use the GPS coordinates where you are.
            if key == "geometry" and len(value) > 0:
                geometry = value.split(" ")
                if len(geometry) == 4:
                    attrs["lat"] = geometry[0]
                    attrs["lon"] = geometry[1]
                continue

            if len(attrs["lat"]) == 0:
                continue
            if key is not None and len(key) > 0 and key in attributes:
                attrs[key] = value
                log.debug("Adding attribute %s with value %s" % (key, value))
            else:
                if key in self.multiple:
                    for item in value:
                        if key in item:
                            for entry in item[key].split():
                                vals = self.getValues(key)
                                if entry in vals:
                                    if vals[entry].find("="):
                                        tmp = vals[entry].split("=")
                                        tags[tmp[0]] = tmp[1]
                                else:
                                    tags[entry] = "yes"
                    continue

                if value is not None and value != "no" and value != "unknown":
                    if key == "track" or key == "geoline":
                        refs.append(tag)
                        log.debug("Adding reference %s" % tag)
                    elif len(value) > 0:
                        if self.privateData(key):
                            priv[key] = value
                        else:
                            tags[key] = value
            if len(tags) > 0:
                feature["attrs"] = attrs
                feature["tags"] = tags
            if len(refs) > 0:
                feature["refs"] = refs
            if len(priv) > 0:
                feature["private"] = priv

        return feature


def main():
    """ """
    parser = argparse.ArgumentParser(description="convert CSV from ODK Central to OSM XML")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-y", "--yaml", help="Alternate YAML file")
    parser.add_argument("-x", "--xlsfile", help="Source XLSFile")
    parser.add_argument("-i", "--infile", required=True, help="The input file downloaded from ODK Central")
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(asctime)s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )
        logging.getLogger("urllib3").setLevel(logging.DEBUG)

    if args.yaml:
        csvin = CSVDump(args.yaml)
    else:
        csvin = CSVDump()
    csvin.parseXLS(args.xlsfile)
    osmoutfile = os.path.basename(args.infile.replace(".csv", ".osm"))
    csvin.createOSM(osmoutfile)

    jsonoutfile = os.path.basename(args.infile.replace(".csv", ".geojson"))
    csvin.createGeoJson(jsonoutfile)

    log.debug("Parsing csv files %r" % args.infile)
    data = csvin.parse(args.infile)
    # This OSM XML file only has OSM appropriate tags and values
    for entry in data:
        feature = csvin.createEntry(entry)
        # Sometimes bad entries, usually from debugging XForm design, sneak in
        if len(feature) == 0:
            continue
        if len(feature) > 0:
            if "lat" not in feature["attrs"]:
                log.warning("Bad record! %r" % feature)
                continue
            csvin.writeOSM(feature)
            # This GeoJson file has all the data values
            csvin.writeGeoJson(feature)
            # print("TAGS: %r" % feature['tags'])

    csvin.finishOSM()
    csvin.finishGeoJson()
    log.info("Wrote OSM XML file: %r" % osmoutfile)
    log.info("Wrote GeoJson file: %r" % jsonoutfile)


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

#
#   Copyright (C) 2020, 2021, 2022, 2023 Humanitarian OpenstreetMap Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import argparse
import logging
import os
import re
import sys
from collections import OrderedDict
from datetime import datetime
from pathlib import Path

import xmltodict

# Instantiate logger
log = logging.getLogger(__name__)


def main():
    """This is a program that reads in the ODK Instance file, which is in XML,
    and converts it to an OSM XML file so it can be viewed in an editor.
    """
    parser = argparse.ArgumentParser(description="Convert ODK XML instance file to OSM XML format")
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument("-i", "--instance", required=True, help="The instance file(s) from ODK Collect")
    # parser.add_argument("-o","--outfile", default='tmp.csv', help='The output file for JOSM')
    args = parser.parse_args()

    # if verbose, dump to the terminal
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    xmlfiles = list()
    if args.instance.find("*") >= 0:
        toplevel = Path()
        for dir in toplevel.glob(args.instance):
            if dir.is_dir():
                xml = os.listdir(dir)
                # There is always only one XML file per instance
                full = os.path.join(dir, xml[0])
                xmlfiles.append(full)
    else:
        toplevel = Path(args.instance)
        if toplevel.is_dir():
            # There is always only one XML file per instance
            full = os.path.join(toplevel, os.path.basename(toplevel))
            xmlfiles.append(full + ".xml")

    # print(xmlfiles)

    # These are all generated by Collect, and can be ignored
    rows = list()
    for instance in xmlfiles:
        logging.info("Processing instance file: %s" % instance)
        with open(instance, "rb") as file:
            # Instances are small, read the whole file
            xml = file.read(os.path.getsize(instance))
            doc = xmltodict.parse(xml)
            fields = list()
            tags = dict()
            data = doc["data"]
            for i, j in data.items():
                if j is None or i == "meta":
                    continue
                # print(f"tag: {i} == {j}")
                pat = re.compile("[0-9.]* [0-9.-]* [0-9.]* [0-9.]*")
                if pat.match(str(j)):
                    if i == "warmup":
                        continue
                    gps = j.split(" ")
                    tags["lat"] = gps[0]
                    tags["lon"] = gps[1]
                    continue
                if type(j) == OrderedDict or type(j) == dict:
                    for ii, jj in j.items():
                        pat = re.compile("[0-9.]* [0-9.-]* [0-9.]* [0-9.]*")
                        if pat.match(str(jj)):
                            gps = jj.split(" ")
                            tags["lat"] = gps[0]
                            tags["lon"] = gps[1]
                            continue
                        if jj is None:
                            continue
                        # print(f"tag: {i} == {j}")
                        if type(jj) == OrderedDict or type(jj) == dict:
                            for iii, jjj in jj.items():
                                if jjj is not None:
                                    tags[iii] = jjj
                                    # print(iii, jjj)
                        else:
                            # print(ii, jj)
                            fields.append(ii)
                            tags[ii] = jj
                else:
                    if i[0:1] != "@":
                        tags[i] = j
        rows.append(tags)

    xml = os.path.basename(xmlfiles[0])
    tmp = xml.replace(" ", "").split("_")
    now = datetime.now()
    timestamp = f"_{now.year}_{now.hour}_{now.minute}"
    outfile = tmp[0] + timestamp + ".csv"

    # with open(outfile, "w", newline="") as csvfile:
    #     fields = list()
    #     for row in rows:
    #         for key in row.keys():
    #             if key not in fields:
    #                 fields.append(key)
    #     csv = csv.DictWriter(csvfile, dialect="excel", fieldnames=fields)
    #     csv.writeheader()
    #     for row in rows:
    #         csv.writerow(row)

    print("Wrote: %s" % outfile)


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     This is free software: you can redistribute it and/or modify

#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import locale
import logging
import os
import sqlite3
import sys

import mercantile

# Instantiate logger
log = logging.getLogger(__name__)


class MapTile(object):
    def __init__(
        self,
        x: int = None,
        y: int = None,
        z: int = None,
        filespec: str = None,
        tile: "MapTile" = None,
        suffix="jpg",
    ):
        """This is a simple wrapper around mercantile.tile to associate a
        filespec with the grid coordinates.

        Args:
            x (int): The X index for this tile
            y (int): The Y index for this tile
            z (int): The Z index for this tile if there is one
            filespec (str): The location of this within the map tile cache
            tile (MapTile): Make a copy of this object
            suffix (str): The image suffix, jpg or png usually

        Returns:
            (MapTile): An instance of this object
        """
        if tile:
            self.x = tile.x
            self.y = tile.y
            self.z = tile.z
        else:
            self.x = x
            self.y = y
            self.z = z
        self.blob = None
        self.filespec = None
        if not filespec and self.z:
            self.filespec = f"{self.z}/{self.y}/{self.x}.{suffix}"
        elif filespec:
            self.filespec = filespec
            tmp = filespec.split("/")
            self.z = tmp[0]
            self.x = tmp[2]
            self.y = tmp[1].replace("." + suffix, "")

    def readImage(self, base: str = "./"):
        """Read a map tile out of the disk based map tile cache.

        Args:
            base (str): The top level directory for the map tile cache
        """
        file = f"{base}/{self.filespec}"
        logging.debug("Adding tile image: %s" % file)
        if os.path.exists(file):
            size = os.path.getsize(file)
            file = open(file, "rb")
            self.blob = file.read(size)

    def dump(self):
        """Dump internal data structures, for debugging purposes only."""
        if self.z:
            print("Z: %r" % self.z)
        if self.x:
            print("X: %r" % self.x)
        if self.y:
            print("Y: %r" % self.y)
        print("Filespec: %s" % self.filespec)
        if self.blob:
            print("Tile size is: %d" % len(self.blob))


class DataFile(object):
    def __init__(
        self,
        dbname: str = None,
        suffix: str = "jpg",
    ):
        """Handle the sqlite3 database file.

        Args:
            dbname (str): The name of the output sqlite file
            suffix (str): The image suffix, jpg or png usually

        Returns:
            (DataFile): An instance of this class
        """
        self.db = None
        self.cursor = None
        if dbname:
            self.createDB(dbname)
        self.dbname = dbname
        self.metadata = None
        self.toplevel = None
        self.suffix = suffix

    def addBounds(
        self,
        bounds: tuple[float, float, float, float],
    ):
        """Mbtiles has a bounds field, Osmand doesn't.

        Args:
            bounds (int): The bounds value for ODK Collect mbtiles
        """
        entry = str(bounds)
        entry = entry[1 : len(entry) - 1].replace(" ", "")
        self.cursor.execute(f"INSERT INTO metadata (name, value) VALUES('bounds', '{entry}')")
        # self.cursor.execute(f"INSERT INTO metadata (name, value) VALUES('minzoom', '9')")
        # self.cursor.execute(f"INSERT INTO metadata (name, value) VALUES('maxzoom', '15')")

    def createDB(
        self,
        dbname: str,
    ):
        """Create and sqlitedb in either mbtiles or Osman sqlitedb format.

        Args:
            dbname (str): The filespec of the sqlite output file
        """
        suffix = os.path.splitext(dbname)[1]

        if os.path.exists(dbname):
            os.remove(dbname)

        self.db = sqlite3.connect(dbname)
        self.cursor = self.db.cursor()
        if suffix == ".mbtiles":
            self.cursor.execute("CREATE TABLE tiles (zoom_level integer, tile_column integer, tile_row integer, tile_data blob)")
            self.cursor.execute("CREATE INDEX tiles_idx on tiles (zoom_level, tile_column, tile_row)")
            self.cursor.execute("CREATE TABLE metadata (name text, value text)")
            # These get populated later
            name = dbname
            description = "Created by osm_fieldwork/basemapper.py"
            self.cursor.execute("CREATE UNIQUE INDEX metadata_idx  ON metadata (name)")
            self.cursor.execute("INSERT INTO metadata (name, value) VALUES('version', '1.1')")
            self.cursor.execute("INSERT INTO metadata (name, value) VALUES('type', 'baselayer')")
            self.cursor.execute(f"INSERT INTO metadata (name, value) VALUES('name', '{name}')")
            self.cursor.execute(f"INSERT INTO metadata (name, value) VALUES('description', '{description}')")
            # self.cursor.execute(f"INSERT INTO metadata (name, value) VALUES('bounds', '{bounds}')")
            self.cursor.execute("INSERT INTO metadata (name, value) VALUES('format', 'jpg')")
        if "sqlite" in suffix:
            # s is always 0
            self.cursor.execute("CREATE TABLE tiles (x int, y int, z int, s int, image blob, PRIMARY KEY (x,y,z,s));")
            self.cursor.execute("CREATE INDEX IND on tiles (x,y,z,s)")
            # Info is simple "2|4" for example, it gets populated later
            self.cursor.execute("CREATE TABLE info (maxzoom Int, minzoom Int);")
            # the metadata is the locale as a string
            loc = locale.getlocale()[0]
            self.cursor.execute(f"CREATE TABLE  android_metadata ({loc})")
        self.db.commit()
        logging.info("Created database file %s" % dbname)

    def writeTiles(
        self,
        tiles: list,
        base: str = "./",
    ):
        """Write map tiles into the to the map tile cache.

        Args:
            tiles (list): The map tiles to write to the map tile cache
            base (str): The default local to write tiles to disk
        """
        for tile in tiles:
            xyz = MapTile(tile=tile)
            xyz.readImage(base)
            # xyz.dump()
            self.writeTile(xyz)

    def writeTile(
        self,
        tile: MapTile,
    ):
        """Write a map tile into the sqlite database file.

        Args:
            tile (MapTile): The map tile to write to the file
        """
        if tile.blob is None:
            logging.error(f"Map tile {tile.filespec} has no image data!")
            # tile.dump()
            return False

        suffix = os.path.splitext(self.dbname)[1]

        if "sqlite" in suffix:
            # Osmand tops out at zoom level 16, so the real zoom level is inverse,
            # and can go negative for really high zoom levels.
            z = 17 - tile.z
            self.db.execute(
                "INSERT INTO tiles (x, y, z, s, image) VALUES (?, ?, ?, ?, ?)",
                [tile.x, tile.y, z, 0, sqlite3.Binary(tile.blob)],
            )

        if suffix == ".mbtiles":
            y = (1 << tile.z) - tile.y - 1
            self.db.execute(
                "INSERT INTO tiles (tile_row, tile_column, zoom_level, tile_data) VALUES (?, ?, ?, ?)",
                [y, tile.x, tile.z, sqlite3.Binary(tile.blob)],
            )

        self.db.commit()


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    parser = argparse.ArgumentParser(description="Create an mbtiles basemap for ODK Collect")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-d", "--database", default="test.mbtiles", help="Database file name")
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    outfile = DataFile(args.database, "jpg")
    toplevel = "/var/www/html/topotiles/"
    foo = "15/12744/6874.png"
    tmp = foo.split("/")
    z = tmp[0]
    x = tmp[1]
    y = tmp[2].replace(".jpg", "")

    # file = "10/388/212.jpg"
    # tile1 = MapTile(x=x, y=y, z=z)
    # tile2 = MapTile(filespec=file)
    # tile2.readImage(f'{toplevel}/{foo}')
    # outfile.writeTile(tile2)

    tile3 = mercantile.Tile(388, 211, 10)
    xyz = MapTile(tile=tile3)
    xyz.readImage(toplevel)
    xyz.dump()



#!/usr/bin/python3

# Copyright (c) 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import json
import logging

# import pandas as pd
import re
import sys
from pathlib import Path

import flatdict
import geojson
import shapely
from geojson import Feature, FeatureCollection, Point, dump

from osm_fieldwork.convert import Convert
from osm_fieldwork.osmfile import OsmFile

log = logging.getLogger(__name__)


class JsonDump(Convert):
    """A class to parse the JSON files from ODK Central or odk2geojson."""

    def __init__(
        self,
        yaml: str = None,
    ):
        """A class to convert the JSON file from ODK Central, or the GeoJson
        file created by the odk2geojson utility.

        Args:
            yaml (str): The filespec of the YAML config file

        Returns:
            (JsonDump): An instance of this object
        """
        self.fields = dict()
        self.nodesets = dict()
        self.data = list()
        self.osm = None
        self.json = None
        self.features = list()
        self.config = super().__init__(yaml)

    # FIXME: a work in progress
    # def parseXLS(self, xlsfile: str):
    #     """Parse the source XLSFile if available to look for details we need"""
    #     if xlsfile is not None and len(xlsfile) > 0:
    #         entries = pd.read_excel(xlsfile, sheet_name=[0])
    #         # There will only be a single sheet
    #         names = entries[0]['name']
    #         defaults = entries[0]['default']
    #         total = len(names)
    #         i = 0
    #         while i < total:
    #             entry = defaults[i]
    #             if str(entry) != 'nan':
    #                 pat = re.compile("..last-saved.*")
    #                 if pat.match(entry):
    #                     name = entry.split('#')[1][:-1]
    #                     self.saved[name] = None
    #                 else:
    #                     self.defaults[names[i]] = entry
    #             i += 1
    #     return True

    def createOSM(
        self,
        filespec: str = "tmp.osm",
    ):
        """Create an OSM XML output files.

        Args:
            filespec (str): The filespec for the output OSM XML file

        Returns:
            (OsmFile): An instance of the OSM XML output file
        """
        log.debug(f"Creating OSM XML file: {filespec}")
        self.osm = OsmFile(filespec)
        return self.osm

    def writeOSM(
        self,
        feature: dict,
    ):
        """Write a feature to an OSM XML output file.

        Args:
            feature (dict): The feature to write to the OSM XML output file
        """
        out = ""
        if "id" in feature["tags"]:
            feature["id"] = feature["tags"]["id"]
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        if "user" in feature["tags"] and "user" not in feature["attrs"]:
            feature["attrs"]["user"] = feature["tags"]["user"]
            del feature["tags"]["user"]
        if "uid" in feature["tags"] and "uid" not in ["attrs"]:
            feature["attrs"]["uid"] = feature["tags"]["uid"]
            del feature["tags"]["uid"]
        if "refs" not in feature:
            out += self.osm.createNode(feature, True)
        else:
            out += self.osm.createWay(feature, True)
        self.osm.write(out)

    def finishOSM(self):
        """Write the OSM XML file footer and close it. The destructor in the
        OsmFile class should do this, but this is the manual way.
        """
        self.osm.footer()

    def createGeoJson(
        self,
        file="tmp.geojson",
    ):
        """Create a GeoJson output file.

        Args:
                file (str): The filespec of the output GeoJson file
        """
        log.debug("Creating GeoJson file: %s" % file)
        self.json = open(file, "w")

    def writeGeoJson(
        self,
        feature: dict,
    ):
        """Write a feature to a GeoJson output file.

        Args:
            feature (dict): The feature to write to the GeoJson output file
        """
        # These get written later when finishing , since we have to create a FeatureCollection
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        self.features.append(feature)

    def finishGeoJson(self):
        """Write the GeoJson FeatureCollection to the output file and close it."""
        features = list()
        for item in self.features:
            # poi = Point()
            poi = Point((float(item["attrs"]["lon"]), float(item["attrs"]["lat"])))
            if "private" in item:
                props = {**item["tags"], **item["private"]}
            else:
                props = item["tags"]
            features.append(Feature(geometry=poi, properties=props))
        collection = FeatureCollection(features)
        dump(collection, self.json)

    def parse(
        self,
        filespec: str = None,
        data: str = None,
    ):
        """Parse the JSON file from ODK Central and convert it to a data structure.
        The input is either a filespec to open, or the data itself.

        Args:
            filespec (str): The JSON or GeoJson input file to convert
            data (str): The data to convert

        Returns:
            (list): A list of all the features in the input file
        """
        log.debug(f"Parsing JSON file {filespec}")
        all_tags = list()
        if not data:
            file = open(filespec, "r")
            infile = Path(filespec)
            if infile.suffix == ".geojson":
                reader = geojson.load(file)
            elif infile.suffix == ".json":
                reader = json.load(file)
            else:
                log.error("Need to specify a JSON or GeoJson file!")
                return all_tags
        elif isinstance(data, str):
            reader = geojson.loads(data)
        elif isinstance(data, list):
            reader = data

        total = list()
        # JSON files from Central use value as the keyword, whereas
        # GeoJSON uses features for the same thing.
        if "value" in reader:
            data = reader["value"]
        elif "features" in reader:
            data = reader["features"]
        else:
            data = reader
        for row in data:
            # log.info(f"ROW: {row}")
            tags = dict()
            if "geometry" in row:
                # If geom not point, convert to centroid
                if row["geometry"]["type"] != "Point":
                    log.debug(f"Converting {row['geometry']['type']} geometry to centroid point")
                    geom = shapely.from_geojson(str(row))
                    centroid = shapely.to_geojson(geom.centroid)
                    row["geometry"] = centroid
                tags["geometry"] = row["geometry"]
            else:
                pat = re.compile("[-0-9.]*, [0-9.-]*, [0-9.]*")
                gps = re.findall(pat, str(row))
                # If geopoint warmup is used, there will be two matches, we only
                # want the second one, which is the location.
                for coords in gps:
                    tags["geometry"] = coords
            if "properties" in row:
                row["properties"]  # A GeoJson formatted file
            else:
                pass  # A JOSM file from ODK Central

            # flatten all the groups into a single data structure
            flattened = flatdict.FlatDict(row)
            for k, v in flattened.items():
                last = k.rfind(":") + 1
                key = k[last:]
                # log.debug(f"Processing tag {key} = {v}")
                # names and comments may have spaces, otherwise
                # it's from a select_multiple
                pat = re.compile("name[:a-z]*")
                names = re.findall(pat, key)
                if len(names) > 0:
                    for name in names:
                        tags[name] = v
                    continue
                if key == "comment":
                    tags[key] = v
                # a JSON file from ODK Central always uses coordinates as
                # the keyword
                if key == "coordinates":
                    if isinstance(v, list):
                        lat = v[1]
                        lon = v[0]
                        tags["geometry"] = f"{lat} {lon}"
                    continue
                if key == "xlocation":
                    tags["geometry"] = v
                    continue
                tags[key] = v
            total.append(tags)

        # log.debug(f"Finished parsing JSON file {filespec}")
        return total

    def createEntry(
        self,
        entry: dict,
    ):
        """Create the feature data structure for this entry.

        Args:
            entry (dict): The feature to convert to the output format

        Returns:
            (dict): The new entry for the output file
        """
        # print(line)
        feature = dict()
        attrs = dict()
        tags = dict()
        priv = dict()
        refs = list()

        # log.debug("Creating entry")
        # First convert the tag to the approved OSM equivalent
        for key, value in entry.items():
            # When using existing OSM data, there's a special geometry field.
            # Otherwise use the GPS coordinates where you are.
            lat = None
            lon = None
            if isinstance(value, float):
                continue
            # log.debug(f"FIXME: {key} = {value} {type(value)}")
            if key == "xid" and value is not None:
                attrs["id"] = int(value)
            if key == "geometry":
                # The GeoJson file has the geometry field. Usually it's a list
                # but on occasion it's a string instead, so turn it into a list
                if isinstance(value, str) and len(coords := value.split(" ")) >= 2:
                    lat = coords[0]
                    lon = coords[1]

                # Parse as geojson
                else:
                    geom = shapely.from_geojson(str(value))

                    if geom.geom_type != "Point":
                        # Use centroid if polygon
                        geom = geom.centroid

                    # Get coords from point
                    lat = geom.y
                    lon = geom.x

                attrs["lat"] = lat
                attrs["lon"] = lon
                # log.debug(f"ATTRS: {attrs}")

            # Some tags are actually attributes
            # print(f"FIXME: {key} {key in attributes}")
            # if key in self.multiple:
            #     for item in value:
            #         if key in item:
            #             for entry in item[key].split():
            #                 vals = self.getValues(key)
            #                 if entry in vals:
            #                     if vals[entry].find("="):
            #                         tmp = vals[entry].split("=")
            #                         tags[tmp[0]] = tmp[1]
            #                 else:
            #                     tags[entry] = "yes"
            #     continue

            if isinstance(value, str) and (value == "no" or value == "unknown"):
                pass
            elif value is not None:
                if key == "track" or key == "geoline":
                    refs.append(tag)
                    log.debug("Adding reference %s" % tag)
                elif len(str(value)) > 0:
                    if self.privateData(key):
                        priv[key] = value
                    else:
                        item = self.convertEntry(key, value)
                        if item is not None and isinstance(item, dict):
                            tags.update(item)
                        elif isinstance(item, list):
                            for entry in item:
                                tags.update(entry)

            if len(tags) > 0:
                if "geometry" in tags:
                    del tags["geometry"]
                feature["attrs"] = attrs
                feature["tags"] = tags
            if len(refs) > 0:
                feature["refs"] = refs
            if len(priv) > 0:
                feature["private"] = priv

        return feature


def json2osm(input_file, yaml_file=None):
    """Process the JSON file from ODK Central or the GeoJSON file to OSM XML format.

    Args:
        input_file (str): The path to the input JSON or GeoJSON file.
        yaml_file (str): The path to the YAML config file (optional).

    Returns:
        osmoutfile (str): Path to the converted OSM XML file.
    """
    log.info(f"Converting JSON file to OSM: {input_file}")
    if yaml_file:
        jsonin = JsonDump(yaml_file)
    else:
        jsonin = JsonDump()

    # jsonin.parseXLS(args.xlsfile)

    # Modify the input file name for the 2 output files, which will get written
    # to the current directory.

    base = Path(input_file).stem
    osmoutfile = f"{base}-out.osm"
    jsonin.createOSM(osmoutfile)

    data = jsonin.parse(input_file)
    # This OSM XML file only has OSM appropriate tags and values

    for entry in data:
        feature = jsonin.createEntry(entry)

        # Sometimes bad entries, usually from debugging XForm design, sneak in
        if len(feature) == 0:
            continue

        if len(feature) > 0:
            if "lat" not in feature["attrs"]:
                if "geometry" in feature["tags"]:
                    if isinstance(feature["tags"]["geometry"], str):
                        coords = list(feature["tags"]["geometry"])
                        # del feature['tags']['geometry']
                elif "coordinates" in feature["tags"]:
                    coords = feature["tags"]["coordinates"]
                    feature["attrs"] = {"lat": coords[1], "lon": coords[0]}
                else:
                    log.warning(f"Bad record! {feature}")
                    continue  # Skip bad records

            jsonin.writeOSM(feature)
    # log.debug("Writing final OSM XML file...")

    # jsonin.finishOSM()
    log.info(f"Wrote OSM XML file: {osmoutfile}")

    return osmoutfile


def main():
    """Run conversion directly from the terminal."""
    parser = argparse.ArgumentParser(description="convert JSON from ODK Central to OSM XML")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-y", "--yaml", help="Alternate YAML file")
    parser.add_argument("-x", "--xlsfile", help="Source XLSFile")
    parser.add_argument("-i", "--infile", required=True, help="The input file downloaded from ODK Central")
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )
        logging.getLogger("urllib3").setLevel(logging.DEBUG)

    json2osm(args.infile, args.yaml)


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()
